import numpy as np
import os
from cartpole_test import sim_rollout, augmented_state

# Global variables
NUM_TRAINING_EPISODES = 1000
NUM_TEST_EPISODES = 100
NUM_DATAPOINTS_PER_EPISODE = 50
DELTA_T = 0.05
SAVE_FOLDER = './data'
VISUALIZE = False
LOG_FREQ = 10
rng = np.random.RandomState(12345)


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    plt.style.use('ggplot')
    from cartpole_sim import CartpoleSim
    from policy import SwingUpAndBalancePolicy, RandomPolicy
    from visualization import Visualizer
    import cv2

    vis = Visualizer(cartpole_length=1.5, x_lim=(0.0, DELTA_T * NUM_DATAPOINTS_PER_EPISODE))
    swingup_policy = SwingUpAndBalancePolicy('policy.npz')
    random_policy = RandomPolicy(seed=12831)
    sim = CartpoleSim(dt=DELTA_T)

    # Initial training data used to train GP for the first episode
    init_state = np.array([0.01, 0.01, np.pi * 0.5, 0.1]) * rng.randn(4)
    ts, state_traj, action_traj = sim_rollout(sim, random_policy, NUM_DATAPOINTS_PER_EPISODE, DELTA_T, init_state)
    delta_state_traj = state_traj[1:] - state_traj[:-1]


    for (split, num_episodes) in [('train', NUM_TRAINING_EPISODES), ('test', NUM_TEST_EPISODES)]:
        print("Generating: %d episodes for %s ... " % (num_episodes, split))
        
        save_path = os.path.join(SAVE_FOLDER, split)
        if not os.path.exists(save_path):
            os.makedirs(save_path)

        for episode in range(num_episodes):
            if episode % LOG_FREQ == 0:
                print("Episode: %d" % episode)
            
            vis.clear()

            # Use learned policy every 4th episode
            if (episode + 1) % 4 == 0:
                policy = swingup_policy
                init_state = np.array([0.01, 0.01, 0.05, 0.05]) * rng.randn(4)
            else:
                policy = random_policy
                init_state = np.array([0.01, 0.01, np.pi * 0.5, 0.1]) * rng.randn(4)

            ts, state_traj, action_traj = sim_rollout(sim, policy, NUM_DATAPOINTS_PER_EPISODE, DELTA_T, init_state)
            delta_state_traj = state_traj[1:] - state_traj[:-1]

            data_img, data_aug_state = [], []
            for i in range(len(state_traj) - 1):
                vis.set_gt_cartpole_state(state_traj[i][3], state_traj[i][2])
                vis.set_gt_delta_state_trajectory(ts[:i+1], delta_state_traj[:i+1])

                if policy == swingup_policy:
                    policy_type = 'swing up'
                else:
                    policy_type = 'random'

                vis_img = vis.draw_gt_cartpole()
                aug_state = np.array(augmented_state(state_traj[i], action_traj[i]))

                data_img.append(vis_img)
                data_aug_state.append(aug_state)

                if VISUALIZE:
                    cv2.imshow('vis', vis_img)
                    cv2.waitKey(int(1000 * DELTA_T))

            #TODO: modify the data saver to suit your modeling needs
            data_img = np.array(data_img)
            data_aug_state = np.array(data_aug_state)
            aug_init_state = np.array(augmented_state(init_state, 0))

            np.savez_compressed(
                "%s/episode_%d" % (save_path, episode),
                img=data_img,
                init_state=aug_init_state,
                aug_state=data_aug_state,
                policy_type=policy_type
            )