import os
import sys

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader, Dataset

import numpy as np


LOG_ITER = 10


class DynamicsDataset(Dataset):
	def __init__(self, data_dir='./data', split='train'):
		self.data_dir = data_dir
		self.split = split
		self.episode_files = os.listdir(os.path.join(self.data_dir, self.split))
		print("Dataset: %s, split: %s, size: %d" % (self.data_dir, 
			self.split, len(self.episode_files)))

	def __len__(self):
		return len(self.episode_files)

	def __getitem__(self, idx):

		# TODO: Modify the dataloader to preprocess and batch your input 
		episode_file = os.path.join(self.data_dir, self.split, 'episode_%d.npz' % idx)
		with np.load(episode_file) as episode:
			images = np.transpose(episode['img'], (0, 3, 1, 2)) / 255.0
			target_states = episode['aug_state']
			init_state = episode['init_state']
		return images, target_states



class StateEstimator(nn.Module):
	def __init__(self):
		super(StateEstimator, self).__init__()

		# TODO: Replace this with your CNN and RNN
		self.model = nn.Identity()

	def forward(self, x):
		x = self.model(x)
		return x



def compute_error(pred, target):
	x_err = np.linalg.norm(target[0] - pred[0])
	dx_err = np.linalg.norm(target[1] - pred[1])
	dtheta_err = np.linalg.norm(target[2] - pred[2])
	sin_theta_err = np.linalg.norm(target[3] - pred[3]) % np.pi
	cos_theta_err = np.linalg.norm(target[4] - pred[4]) % np.pi

	return x_err, dx_err, dtheta_err, sin_theta_err, cos_theta_err


def train(train_dataloader, test_dataloader):
	model = StateEstimator()
	loss_fn = nn.SmoothL1Loss(reduction='mean') # TODO: Replace this with your loss
	optimizer = torch.optim.SGD(model.parameters(), lr=0.001, momentum=0.9)

	for epoch in range(20):

		# Training Loop
		print("Epoch: %d" % epoch)
		for batch_idx, batch in enumerate(train_dataloader):
			images, target_states = batch
			optimizer.zero_grad()

			images = images.reshape(-1, *images.shape[2:]).float()
			target = target_states.reshape(-1, *target_states.shape[2:])[:, :5].float()

			# TODO: Replace this with your forward and backward updates
			pred = model(images)
			loss = loss_fn(pred, target)
			loss.backward()
			optimizer.step()

			pred = pred.detach().cpu().numpy()
			x_err, dx_err, dtheta_err, sin_theta_err, cos_theta_err = compute_error(pred, target)

			if batch_idx % LOG_ITER == 0:
				print("Training Batch (%d Iters) - Loss: %f, x err: %f, dx err: %f, dtheta_err: %f, sin_theta_err %f, cos_theta_err: %f" % 
					  (batch_idx, loss.item(), x_err, dx_err, dtheta_err, sin_theta_err, cos_theta_err))


		# Evaluation Loop
		print("Eval Run: %d" % epoch)
		x_err_acc, dx_err_acc, dtheta_err_acc, sin_theta_err_acc, cos_theta_err_acc = [], [], [], [], []
		for batch_idx, batch in enumerate(test_dataloader):
			images, target_states = batch

			images = images.reshape(-1, *images.shape[2:]).float()
			target = target_states.reshape(-1, *target_states.shape[2:])[:, :5].float()

			with torch.no_grad():
				pred = model(images)

			pred = pred.detach().cpu().numpy()
			x_err, dx_err, dtheta_err, sin_theta_err, cos_theta_err = compute_error(pred, target)
			x_err_acc.append(x_err)
			dx_err_acc.append(dx_err)
			dtheta_err_acc.append(dtheta_err)
			sin_theta_err_acc.append(sin_theta_err)
			cos_theta_err_acc.append(cos_theta_err)
		print("Eval Run %d, Total Error - x err: %f, dx err: %f, dtheta_err: %f, sin_theta_err %f, cos_theta_err: %f" % 
			 (epoch, np.sum(x_err_acc), np.sum(dx_err_acc), np.sum(dtheta_err_acc), np.sum(sin_theta_err_acc), np.sum(cos_theta_err_acc)))


if __name__ == "__main__":
	batch_size = 1

	train_dataset = DynamicsDataset(data_dir='./data', split='train')
	train_dataloader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

	test_dataset = DynamicsDataset(data_dir='./data', split='test')
	test_dataloader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)

	train(train_dataloader, test_dataloader)